import 'dart:convert';

import 'package:equatable/equatable.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:ulid/ulid.dart';
import 'package:ocg_user/ocg_user.dart';

part 'message.g.dart';

class MessageType {
  static String get text => "text";
  static String get image => "image";
  static String get video => "video";
  static String get file => "file";
  static String get location => "location";
}

class MessageNetworkState {
  static int get failed => -1;
  static int get progress => 0;
  static int get sent => 1;
}

@JsonSerializable(explicitToJson: true)
class Message extends Equatable {
  @JsonKey(name: "msg_id")
  final String id;
  @JsonKey(name: "chat_id")
  final String chatId;
  @JsonKey(name: "msg")
  final Map<String, dynamic> content;
  @JsonKey(name: "sender_id")
  final String senderId;
  @JsonKey(name: "msg_type")
  final String type;
  @JsonKey(name: "reply_to")
  final String replyTo;
  @JsonKey(name: "rand_id")
  final String randId;
  @JsonKey(ignore: true, defaultValue: false)
  final bool isSelf;
  final int state;
  final bool isSeen;
  @JsonKey(ignore: true)
  final Map<String, MessageReceiveData> receiveData;
  @JsonKey(
    ignore: true,
  )
  final User sender;

  final Message replyToMsg;

  Message(
    this.id,
    this.chatId,
    this.content,
    this.senderId,
    this.type,
    this.replyTo,
    this.randId,
  )   : this.isSelf = false,
        this.sender = null,
        this.state = MessageNetworkState.progress,
        this.isSeen = false,
        this.receiveData = {},
        this.replyToMsg = null;

  Message.named(
      {this.id,
      this.chatId,
      this.content,
      this.isSelf,
      this.senderId,
      this.sender,
      this.type,
      this.replyTo,
      this.isSeen,
      this.state,
      this.receiveData,
      this.replyToMsg,
      this.randId});

  String get text {
    if (type == MessageType.text) {
      return TextMessage.fromJson(content).content;
    }
    if (type == MessageType.image) {
      return "📷 Image";
    }
    if (type == MessageType.video) {
      return "📹 Video";
    }
    if (type == MessageType.file) {
      return "📁 File";
    }
    if (type == MessageType.location) {
      return "Location";
    }
    return "Start Conversation";
  }

  bool get isSent => state == MessageNetworkState.sent;

  bool get isFailed => state == MessageNetworkState.failed;

  bool get isProgress => state == MessageNetworkState.progress;

  Map<String, dynamic> toJson() => <String, dynamic>{
        'msg_id': id,
        'chat_id': chatId,
        'msg': jsonEncode(content),
        'sender_id': senderId,
        'msg_type': type,
        "reply_to": replyTo,
        "rand_id": randId,
      };

  factory Message.fromJson(json) => Message.named(
      id: json['msg_id'] as String,
      chatId: json['chat_id'] as String,
      content: (json['msg'] as String) != null ? jsonDecode(json["msg"]) : {},
      senderId: json['sender_id'] as String,
      type: json['msg_type'] as String,
      replyToMsg:
          json["reply_to"] != null ? Message.fromJson(json["reply_to"]) : null,
      randId: json['rand_id'] as String);

  Message copyWith(
      {String id,
      String chatId,
      Map<String, dynamic> content,
      bool isSelf,
      int state,
      String senderId,
      User sender,
      String type,
      String replyTo,
      DateTime recordedAt,
      List<String> readBy,
      bool isSeen,
      List<String> deliveredTo,
      replyToMsg,
      String randId}) {
    return Message.named(
        id: id ?? this.id,
        chatId: chatId ?? this.chatId,
        content: content ?? this.content,
        isSelf: isSelf ?? this.isSelf,
        sender: sender ?? this.sender,
        senderId: senderId ?? this.senderId,
        replyTo: replyTo ?? this.replyTo,
        type: type ?? this.type,
        isSeen: isSeen ?? this.isSeen,
        receiveData: receiveData ?? this.receiveData,
        state: state ?? this.state,
        replyToMsg: replyToMsg ?? this.replyToMsg,
        randId: randId ?? this.randId);
  }

  @override
  List<Object> get props => [id];

  DateTime get recordedAt => id != null
      ? DateTime.fromMillisecondsSinceEpoch(Ulid.parse(id).toMillis())
      : DateTime.now();
}

@JsonSerializable(explicitToJson: true)
class MessageReceiveData {
  final String type;
  final DateTime recordedAt;

  MessageReceiveData(this.type, this.recordedAt);

  MessageReceiveData.named({this.type, this.recordedAt});

  Map<String, dynamic> toJson() => _$MessageReceiveDataToJson(this);

  factory MessageReceiveData.fromJson(json) =>
      _$MessageReceiveDataFromJson(json);
}

abstract class MediaMessage {}

@JsonSerializable(explicitToJson: true)
class TextMessage extends MediaMessage {
  final String content;

  TextMessage(this.content);

  TextMessage.named({this.content});

  Map<String, dynamic> toJson() => _$TextMessageToJson(this);

  factory TextMessage.fromJson(json) => _$TextMessageFromJson(json);
}

@JsonSerializable(explicitToJson: true)
class ImageMessage extends MediaMessage {
  final String url;

  ImageMessage(this.url);

  Map<String, dynamic> toJson() => _$ImageMessageToJson(this);

  factory ImageMessage.fromJson(json) => _$ImageMessageFromJson(json);
}

@JsonSerializable(explicitToJson: true)
class VideoMessage {
  final String url;
  final String thumbnail;

  VideoMessage(this.url, this.thumbnail);

  VideoMessage.named({this.url, this.thumbnail});

  Map<String, dynamic> toJson() => _$VideoMessageToJson(this);

  factory VideoMessage.fromJson(json) => _$VideoMessageFromJson(json);
}

@JsonSerializable(explicitToJson: true)
class FileMessage {
  final String url;
  final String name;
  final int size;

  FileMessage(this.url, this.name, this.size);

  FileMessage.named({this.url, this.name, this.size});

  Map<String, dynamic> toJson() => _$FileMessageToJson(this);

  factory FileMessage.fromJson(json) => _$FileMessageFromJson(json);
}

@JsonSerializable(explicitToJson: true)
class LocationMessage {
  final String placeId;
  final String fullAddress;
  final double latitude;
  final double longitude;

  LocationMessage(
      this.placeId, this.fullAddress, this.latitude, this.longitude);

  LocationMessage.named(
      {this.placeId, this.fullAddress, this.latitude, this.longitude});

  Map<String, dynamic> toJson() => _$LocationMessageToJson(this);

  factory LocationMessage.fromJson(json) => _$LocationMessageFromJson(json);
}

@JsonSerializable(explicitToJson: true)
class MessageResponse extends Equatable {
  @JsonKey(name: "msg_id")
  final String msgId;
  @JsonKey(name: "rand_id")
  final String randId;

  MessageResponse(this.msgId, this.randId);

  MessageResponse.named(this.msgId, this.randId);

  Map<String, dynamic> toJson() => _$MessageResponseToJson(this);

  factory MessageResponse.fromJson(json) => _$MessageResponseFromJson(json);

  @override
  List<Object> get props => [msgId, randId];
}
