import 'dart:async';
import 'dart:convert';

import 'package:ocg_app/ocg_app.dart';
import 'package:flutter/foundation.dart';
import 'package:ocg_user/ocg_user.dart';
import 'package:http/http.dart' as http;

import '../models/chat.dart';
import '../models/signal.dart';

abstract class ChatRepository {
  Future<AppResponse<String>> createChat(
    Chat chat,
  );
  Future<AppResponse<String>> startChat(
    String friendId,
  );
  Future<AppResponse<List<Chat>>> getChats();
  Future<AppResponse<bool>> addSignal(ChatSignal signal);
  Stream<ChatSignal> listenSignals();
  Future<AppResponse<List<Chat>>> getGroupsInCommon(String userId);
  Future<AppResponse<Chat>> getChatDetails(String chatId, {String userId});
  Future<AppResponse<bool>> renameGroup(String chatId, String name);
  Future<AppResponse<bool>> removeGroupUser(String chatId, String userId);
  Future<AppResponse<bool>> makeGroupAdmin(String chatId, String userId);
  Future<AppResponse<bool>> addGroupUser(String chatId, List<String> userIds);
  Future<AppResponse<bool>> muteChat(String chatId, {bool mute = true});
  Future<AppResponse<bool>> updateMetadata(
      String chatId, Map<String, dynamic> metadata);
  Future<AppResponse<bool>> leaveGroup(String chatId);
  Future<AppResponse<bool>> clearChat(
    String chatId,
  );
}

class ChatTestRepository extends ChatRepository {
  final UserRepository userRepository;
  final List<Chat> _chats;
  final StreamController<ChatSignal> _streamController;

  ChatTestRepository(this.userRepository)
      : _chats = [],
        _streamController = StreamController<ChatSignal>();
  @override
  Future<AppResponse<bool>> addSignal(ChatSignal signal) async {
    _streamController.add(signal);
    return AppResponse.named(data: true);
  }

  @override
  Future<AppResponse<String>> createChat(
    Chat chat,
  ) async {
    _chats.add(chat.copyWith());
    return AppResponse.named(data: "true");
  }

  @override
  Future<AppResponse<List<Chat>>> getChats() async {
    return AppResponse.named(data: _chats.toList());
  }

  @override
  Stream<ChatSignal> listenSignals() => _streamController.stream;

  @override
  Future<AppResponse<bool>> addGroupUser(String chatId, List<String> userIds) {
    // TODO: implement addGroupUser
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> clearChat(String chatId) {
    // TODO: implement clearChat
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<Chat>> getChatDetails(String chatId, {String userId}) {
    // TODO: implement getChatDetails
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<List<Chat>>> getGroupsInCommon(String userId) {
    // TODO: implement getGroupsInCommon
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> leaveGroup(String chatId) {
    // TODO: implement leaveGroup
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> makeGroupAdmin(String chatId, String userId) {
    // TODO: implement makeGroupAdmin
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> muteChat(String chatId, {bool mute = true}) {
    // TODO: implement muteChat
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> removeGroupUser(String chatId, String userId) {
    // TODO: implement removeGroupUser
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> renameGroup(String chatId, String name) {
    // TODO: implement renameGroup
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<String>> startChat(String friendId) {
    // TODO: implement startChat
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> updateMetadata(
      String chatId, Map<String, dynamic> metadata) {
    // TODO: implement updateMetadata
    throw UnimplementedError();
  }
}

class ChatNetworkRepository extends ChatRepository {
  final UserRepository userRepository;
  final http.BaseClient client;

  ChatNetworkRepository(
    this.client,
    this.userRepository,
  );
  @override
  Future<AppResponse<bool>> addGroupUser(
      String chatId, List<String> userIds) async {
    try {
      await Future.wait(userIds.map((id) async {
        await _addGroupSingleUser(chatId, id);
      }));
      return AppResponse.named(data: true);
    } catch (e) {
      debugPrint(
        "Error $e",
      );
      return AppResponse.named(error: e.toString());
    }
  }

  Future<AppResponse<bool>> _addGroupSingleUser(
      String chatId, String userId) async {
    try {
      String url = "/chats/$chatId/users/$userId";
      var response = await client.post(
        url,
      );
      return AppResponse.named(data: response.statusCode == 200);
    } catch (e) {
      debugPrint(
        "Error $e",
      );
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> addSignal(ChatSignal signal) {
    // TODO: implement addSignal
    throw UnimplementedError();
  }

  @override
  Stream<ChatSignal> listenSignals() {
    // TODO: implement listenSignals
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> clearChat(String chatId) async {
    try {
      String url = "/chats/$chatId/clear";
      var response = await client.post(url, body: jsonEncode({}));
      return AppResponse.named(data: response.statusCode == 200);
    } catch (e) {
      debugPrint(
        "Error $e",
      );
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<String>> createChat(Chat chat) async {
    try {
      String url = "/chats/";
      var response = await client.post(url,
          body: jsonEncode({"with": chat.users, "name": chat.groupName}),
          headers: ApiConfig.headers);
      String chatId = jsonDecode(response.body)["chat_id"];
      if (chatId != null && chat.metadata != null) {
        AppResponse<bool> metaResponse =
            await updateMetadata(chatId, chat.metadata);
        if (metaResponse.isSuccess) {}
      }
      return AppResponse.named(data: chatId);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<Chat>> getChatDetails(String chatId,
      {String userId}) async {
    try {
      AppResponse<List<Chat>> chatResponse = await getChats();
      if (!chatResponse.isSuccess) {
        return AppResponse.named(error: "Chat not found");
      }
      int index = -1;
      List<Chat> chats = chatResponse.data;
      index = chats.indexWhere((val) => val.id == chatId);
      if (userId != null) {
        index = chats
            .indexWhere((val) => !val.isGroup && val.users.contains(userId));
      }
      if (index != -1) {
        var chat = chats[index];
        if (!(chat.isGroup ?? true)) {
          AppResponse<String> userIdResponse = await userRepository.getSelfId();
          if (!userIdResponse.isSuccess) {
            return null;
          }
          String selfId = userIdResponse.data;
          List<String> users =
              chat.users.where((element) => selfId != element).toList();
          if (users.length == 1) {
            AppResponse<User> userResponse =
                await userRepository.getUser(users[0]);
            if (userResponse.isSuccess) {
              chat = chat.copyWith(friend: userResponse.data);
            }
          }
        }
        return AppResponse.named(data: chat);
      }
      return AppResponse.named(error: "Chat not found");
    } catch (e) {
      debugPrint("Error $e");
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<List<Chat>>> getChats() async {
    try {
      String url = "/chats/";
      var response = await client.get(
        url,
      );

      var json = jsonDecode(response.body);
      if (json["chats"] == null || !(json["chats"] is List)) {
        return AppResponse.named(data: []);
      }
      var chatJson = json["chats"];
      List<Chat> chats = (chatJson as List).map((val) {
        try {
          return Chat.fromJson(val);
        } catch (e) {
          debugPrint("Error $e");
          return null;
        }
      }).toList();

      chats = chats.where((val) => val != null).toList();

      AppResponse<String> userIdResponse = await userRepository.getSelfId();
      if (!userIdResponse.isSuccess) {
        return AppResponse.named(data: []);
      }
      String userId = userIdResponse.data;
      chats = await Future.wait(chats.map((chat) async {
        if (!(chat.isGroup ?? true)) {
          List<String> users =
              chat.users.where((element) => element != userId).toList();
          if (users.length == 1) {
            AppResponse<User> userResponse =
                await userRepository.getUser(users[0]);
            if (userResponse.isSuccess) {
              chat = chat.copyWith(friend: userResponse.data);
            }
          }
        }
        return chat;
      }).toList());

      return AppResponse.named(data: chats);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<List<Chat>>> getGroupsInCommon(String userId) async {
    try {
      String url = "/common_chats";
      var response =
          await client.post(url, body: jsonEncode({"user_id": userId}));
      var json = ((jsonDecode(response.body))["result"]) ?? [];
      List<String> chatIds = List<String>.from(json);
      AppResponse<List<Chat>> currentGroupResponse = await getChats();
      if (!currentGroupResponse.isSuccess) {
        return AppResponse.named(error: currentGroupResponse.error);
      }
      return AppResponse.named(
          data: currentGroupResponse.data
              .where((val) => chatIds.contains(val.id))
              .toList());
    } catch (e) {
      debugPrint(
        "Error $e",
      );
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> leaveGroup(String chatId) async {
    try {
      String url = "/chats/$chatId/users/";
      var response = await client.delete(
        url,
      );
      return AppResponse.named(data: response.statusCode == 200);
    } catch (e) {
      debugPrint("Error $e");
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> makeGroupAdmin(String chatId, String userId) async {
    try {
      String url = "/chats/$chatId/make_admin";
      var response =
          await client.post(url, body: jsonEncode({"user_id": userId}));
      debugPrint("Make Admin ${response.body}");
      return AppResponse.named(data: response.statusCode == 200);
    } catch (e) {
      debugPrint("Error $e");
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> muteChat(String chatId, {bool mute = true}) async {
    try {
      String url = "/chats/$chatId/${mute ? "mute" : "unmute"}";
      var response = await client.post(
        url,
      );
      return AppResponse.named(data: response.statusCode == 200);
    } catch (e) {
      debugPrint("Error $e");
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> removeGroupUser(
      String chatId, String userId) async {
    try {
      String url = "/chats/$chatId/users/$userId";
      var response = await client.delete(
        url,
      );
      debugPrint("Rename ${response.body}");
      return AppResponse.named(data: response.statusCode == 200);
    } catch (e) {
      debugPrint("Error $e");

      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> renameGroup(String chatId, String name) async {
    try {
      String url = "/chats/$chatId/rename";
      var response = await client.post(url, body: jsonEncode({"name": name}));
      debugPrint("Rename ${response.body}");
      return AppResponse.named(data: response.statusCode == 200);
    } catch (e) {
      debugPrint("Error $e");

      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<String>> startChat(String friendId) async {
    try {
      String url = "/chats/";
      var response = await client.post(url,
          body: jsonEncode(
            {"with": friendId},
          ));
      return AppResponse.named(data: jsonDecode(response.body)["chat_id"]);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> updateMetadata(
      String chatId, Map<String, dynamic> metadata) async {
    try {
      String url = "/chats/$chatId/metadata";
      var response = await client.post(url,
          body: jsonEncode({"metadata": jsonEncode(metadata)}));
      return AppResponse.named(data: response.statusCode == 200);
    } catch (e) {
      debugPrint(
        "Error $e",
      );
      return AppResponse.named(error: e.toString());
    }
  }
}
