import 'dart:async';
import 'dart:convert';

import 'package:ocg_app/ocg_app.dart';
import 'package:ocg_chat/src/models/chat.dart';
import 'package:ocg_chat/src/models/message.dart';
import 'package:ocg_chat/src/repository/chat.dart';
import 'package:flutter/material.dart';
import 'package:ocg_user/ocg_user.dart';
import 'package:http/http.dart' as http;

abstract class MessageRepository {
  Future<AppResponse<List<Message>>> getSharedFilesList(String chatId);
  Future<AppResponse<bool>> readChat(String chatId, String msgId);
  Future<AppResponse<MessageResponse>> sendMessage(
    Message message,
  );
  Future<AppResponse<List<Message>>> getMessages(String chatId,
      {String before});
  Future<Message> decodeMessage(Message message);
  Future<List<Message>> decodeMessageList(var messagesJson);
  Future<AppResponse<List<Message>>> searchMessages(
      String chatId, String query);
  Future<AppResponse<bool>> deleteMessage(String chatId, String msgId,
      {bool onlyMine = true});
  Future<AppResponse<List<User>>> getReadyBy(String chatId, String messageId);
}

class MessageTestRepository extends MessageRepository {
  final UserRepository userRepository;

  MessageTestRepository(this.userRepository);

  @override
  Future<Message> decodeMessage(Message message) {
    // TODO: implement decodeMessage
    throw UnimplementedError();
  }

  @override
  Future<List<Message>> decodeMessageList(messagesJson) {
    // TODO: implement decodeMessageList
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> deleteMessage(String chatId, String msgId,
      {bool onlyMine = true}) {
    // TODO: implement deleteMessage
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<List<Message>>> getMessages(String chatId,
      {String before}) {
    // TODO: implement getMessages
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<List<User>>> getReadyBy(String chatId, String messageId) {
    // TODO: implement getReadyBy
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<List<Message>>> getSharedFilesList(String chatId) {
    // TODO: implement getSharedFilesList
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<bool>> readChat(String chatId, String msgId) {
    // TODO: implement readChat
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<List<Message>>> searchMessages(
      String chatId, String query) {
    // TODO: implement searchMessages
    throw UnimplementedError();
  }

  @override
  Future<AppResponse<MessageResponse>> sendMessage(Message message) {
    // TODO: implement sendMessage
    throw UnimplementedError();
  }
}

class MessageNetworkRepository extends MessageRepository {
  final UserRepository userRepository;
  final ChatRepository chatRepository;
  final http.BaseClient client;

  MessageNetworkRepository(
    this.client,
    this.userRepository,
    this.chatRepository,
  );
  @override
  Future<Message> decodeMessage(Message oldMessage) async {
    if (oldMessage.replyToMsg != null) {
      Message temp = oldMessage.replyToMsg;
      Message newTemp = await decodeMessage(temp);
      oldMessage = oldMessage.copyWith(replyToMsg: newTemp);
    }
    Message val = oldMessage.copyWith();
    AppResponse<User> userResponse = await userRepository.getUser(val.senderId);
    if (userResponse.isSuccess) {
      var idResponse = await userRepository.getSelfId();
      if (idResponse.isSuccess) {
        val = val.copyWith(
            state: MessageNetworkState.sent,
            sender: userResponse.data,
            isSelf: idResponse.data == val.senderId);
      }
      return val;
    }
    return null;
  }

  @override
  Future<List<Message>> decodeMessageList(messagesJson) async {
    List<Message> messages = List<Message>.from(
        (messagesJson as List).map((val) => Message.fromJson(val)).toList());
    return await Future.wait(messages.map((e) async {
      return await decodeMessage(e);
    }));
  }

  @override
  Future<AppResponse<bool>> deleteMessage(String chatId, String msgId,
      {bool onlyMine = true}) async {
    try {
      String url = "/chats/$chatId/messages/$msgId/delete";
      var response =
          await client.post(url, body: jsonEncode({"only_mine": onlyMine}));
      var json = jsonDecode(response.body);
      if (json["status"] == "ok") {
        return AppResponse.named(data: true);
      }
      return AppResponse.named(error: "Message not deleted");
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<List<Message>>> getMessages(String chatId,
      {String before}) async {
    try {
      String url = "/chats/$chatId/${before != null ? "?before=$before" : ""}";
      AppResponse<Chat> chatResponse =
          await chatRepository.getChatDetails(chatId);
      if (!chatResponse.isSuccess) {
        return AppResponse.named(data: []);
      }
      Chat chat = chatResponse.data;
      var response = await client.get(
        url,
      );
      var json = jsonDecode(response.body);
      if (json["messages"] == null || !(json["messages"] is List)) {
        return AppResponse.named(error: "Messages not exists");
      }
      var messagesJson = json["messages"] as List;
      messagesJson.forEach((element) {
        if (element["replyTo"] != null) {
          debugPrint("Reply $element");
        }
      });
      List<Message> result = List<Message>.from(
          messagesJson.map((val) => Message.fromJson(val)).toList());
      List<Message> decodedMessages = [];
      await Future.wait(result.map((val) async {
        Message message = await decodeMessage(val);
        if (message != null) {
          message = message.copyWith(
              isSeen: (message.isSelf ?? false) &&
                  (chat?.lastReadMegRecordedAt?.compareTo(message.recordedAt) ??
                          0) >=
                      0);
          decodedMessages.add(message);
        }
      }));
      decodedMessages.sort((a, b) => a.recordedAt.compareTo(b.recordedAt));
      decodedMessages.forEach((val) {
        if (chat.lastReadMegRecordedAt.compareTo(val.recordedAt) >= 0) {
          val = val.copyWith(isSeen: true);
        }
      });
      return AppResponse.named(
          data: decodedMessages.where((val) => val != null).toList());
    } catch (e) {
      debugPrint("Message Error $e");
      return AppResponse.named(error: e.toString());
    }
  }

  Future<AppResponse> getSharedMessages(String chatId) async {
    List<Message> messages = [];
    List<String> apis = ["images", "audios", "videos", "files"];
    List<List<Message>> data = await Future.wait(apis.map((e) async {
      try {
        String url = "/chats/$chatId/$e";
        var response = await client.get(
          url,
        );
        var json = (jsonDecode(response.body)["messages"]) ?? [];
        return (json as List).map((val) => Message.fromJson(val)).toList();
      } catch (e) {
        debugPrint(
          "Error $e",
        );
        return [];
      }
    }));
    data.forEach((list) {
      if (list != null) messages.addAll(list);
    });
    return AppResponse.named(data: messages);
  }

  @override
  Future<AppResponse<List<User>>> getReadyBy(
      String chatId, String messageId) async {
    try {
      debugPrint("Loading Read By");
      String url = "/chats/$chatId/messages/$messageId/users";
      var response = await client.get(url);
      var json = jsonDecode(response.body);
      if (json["users"] != null && json["users"] is List) {
        var userIds = json["users"] ?? [];
        var result =
            await Future.wait(List<String>.from(userIds as List).map((e) async {
          AppResponse<User> userResponse = await userRepository.getUser(e);
          if (userResponse.isSuccess) {
            return userResponse.data;
          }
        }).toList());

        result = result.where((element) => element != null).toList();
        return AppResponse.named(data: List<User>.from(result));
      }

      return AppResponse.named(error: "Some error occurred");
    } catch (e) {
      debugPrint("Error ${e.toString()}");

      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<List<Message>>> getSharedFilesList(String chatId) async {
    try {
      List<Message> messages = [];
      List<String> apis = ["images", "audios", "videos", "files"];
      List<List<Message>> data = await Future.wait(apis.map((e) async {
        try {
          String url = "/chats/$chatId/$e";
          var response = await client.get(
            url,
          );
          var json = (jsonDecode(response.body)["messages"]) ?? [];
          return (json as List).map((val) => Message.fromJson(val)).toList();
        } catch (e) {
          debugPrint(
            "Error $e",
          );
          return [];
        }
      }));
      data.forEach((list) {
        if (list != null) messages.addAll(list);
      });
      return AppResponse.named(data: messages);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<bool>> readChat(String chatId, String msgId) async {
    try {
      String url = "/chats/$chatId/read";
      await client.post(url,
          body: jsonEncode({"msg_id": msgId}), headers: ApiConfig.headers);
      return AppResponse.named(data: true);
    } catch (e) {
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<List<Message>>> searchMessages(
      String chatId, String query) async {
    try {
      String url = "/chats/$chatId/search";
      var response = await client.post(url,
          body: jsonEncode({"query": query}), headers: ApiConfig.headers);
      var messagesJson = (jsonDecode(response.body)["messages"]) ?? [];
      List<Message> messages = await decodeMessageList(messagesJson);
      return AppResponse.named(data: messages);
    } catch (e) {
      debugPrint(
        "Error $e",
      );
      return AppResponse.named(error: e.toString());
    }
  }

  @override
  Future<AppResponse<MessageResponse>> sendMessage(Message message) async {
    try {
      String url = "/chats/${message.chatId}/";
      var json = message.toJson();
      json = JsonHelper.cleanJson(json);
      var response = await client.post(url,
          body: jsonEncode(json), headers: ApiConfig.headers);
      var respnseJson = jsonDecode(response.body);
      return AppResponse.named(data: MessageResponse.fromJson(respnseJson));
    } catch (e) {
      debugPrint("Error $e");
      return AppResponse.named(error: e.toString());
    }
  }
}
