import 'package:ocg_chat/ocg_chat.dart';
import 'package:ocg_chat/src/bloc/chat_signal/state.dart';
import 'package:ocg_chat/src/models/notification.dart';
import 'package:ocg_chat/src/models/signal.dart';
import 'package:ocg_chat/src/models/socket.dart';
import 'package:firebase_messaging/firebase_messaging.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:ocg_contact/ocg_contact.dart';
import 'package:ocg_login/ocg_login.dart';
import 'package:ocg_user/ocg_user.dart';

import '../bloc/chat/state.dart';
import '../bloc/chat/bloc.dart';
import '../bloc/chat/event.dart';

import '../bloc/chat_signal/bloc.dart';

import '../models/chat.dart';
import '../routes/routes.dart';

import '../widgets/chat_widget.dart';

class ChatScreen extends StatefulWidget {
  final ChatSocket chatSocket;
  ChatScreen({Key key, this.chatSocket}) : super(key: key);

  @override
  _ChatScreenState createState() => _ChatScreenState();
}

class _ChatScreenState extends State<ChatScreen> {
  @override
  void initState() {
    super.initState();
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    BlocProvider.of<ChatBloc>(context).add(ChatRefreshEvent());
    BlocProvider.of<ContactSyncBloc>(context).add(ContactSyncAllEvent());
    BlocProvider.of<ContactListBloc>(context).add(ContactListRefreshEvent());
    subscribeToFcm();
    subscribeToSocket();
  }

  Future<void> subscribeToSocket() async {
    await widget.chatSocket.connect();
    widget.chatSocket.subscribeToPersonalChat();
    widget.chatSocket.stream.listen((event) {
      if (event is SocketNewChatData || event is SocketMessageData) {
        BlocProvider.of<ChatBloc>(context).add(ChatRefreshEvent());
      }
      if (event is SocketMessageData) {
        BlocProvider.of<MessageBloc>(context)
            .add(MessageReceiveEvent(event.message));
      }
    });
  }

  Future<void> subscribeToFcm() async {
    FirebaseMessaging firebaseMessaging = FirebaseMessaging();
    firebaseMessaging.setAutoInitEnabled(true);
    firebaseMessaging.configure(onLaunch: (data) async {
      debugPrint("Launch");
      // onData(data);
    }, onMessage: (data) async {
      debugPrint("Message");
      // onData(data);
    }, onResume: (data) async {
      debugPrint("Resume");
      onData(data);
    });
  }

  Future<void> onData(ms) async {
    if (ms["data"] != null) {
      if (ms["data"]["type"] == NotificationType.msg) {
        Navigator.of(context).pushNamed(ChatRouteType.messagesScreen,
            arguments: {"chatId": ms["data"]["metaId"]});
      }
    }
  }

  @override
  Widget build(BuildContext context) {
    return WillPopScope(
      onWillPop: () async {
        return false;
      },
      child: AnnotatedRegion<SystemUiOverlayStyle>(
        value: SystemUiOverlayStyle(statusBarBrightness: Brightness.dark),
        child: Scaffold(
          backgroundColor: Colors.white,
          body: BlocListener<LoginBloc, LoginState>(
            listener: (context, state) {
              if (state is LoginExistState) {}
            },
            child: BlocBuilder<LoginBloc, LoginState>(
                builder: (context, loginState) {
              if (!(loginState is LoginExistState)) {
                return Scaffold(
                  body: Center(
                    child: Text("Not exists"),
                  ),
                );
              }
              LoginExistState existState = loginState;
              User user = existState.user;
              return SafeArea(
                child: BlocListener<ChatBloc, ChatState>(
                  listener: (context, state) {},
                  child: BlocBuilder<ChatBloc, ChatState>(
                    builder: (context, state) {
                      if (state is ChatLoadingState) {
                        return Center(child: CircularProgressIndicator());
                      }
                      if (state is ChatLoadedState) {
                        List<Chat> chats = state.chats;
                        return Scaffold(
                          body: BlocBuilder<ChatSignalBloc, ChatSignalState>(
                            builder: (context, signalState) {
                              ChatSignal typingSignal;
                              if (signalState is ChatSignalLoadedState) {
                                int index = signalState.signals.indexWhere(
                                    (e) =>
                                        e.message == ChatSignalType.typing &&
                                        e.senderId != user.id);
                                if (index != -1)
                                  typingSignal = signalState.signals[index];
                              }
                              List<ChatWidget> children =
                                  List<ChatWidget>.from(chats
                                      .map((chat) => ChatWidget(
                                            chat: chat,
                                            metadata: typingSignal != null &&
                                                    typingSignal.channelId ==
                                                        chat.id
                                                ? "${typingSignal.sender?.name} typing"
                                                : null,
                                            onTap: (chat) {
                                              Navigator.of(context).pushNamed(
                                                  ChatRouteType.messagesScreen,
                                                  arguments: {
                                                    "chatId": chat.id
                                                  });
                                            },
                                          ))
                                      .toList());
                              return Scaffold(
                                appBar: AppBar(
                                  title: Text("Chats"),
                                ),
                                body: RefreshIndicator(
                                  onRefresh: () async {
                                    BlocProvider.of<ChatBloc>(context)
                                        .add(ChatRefreshEvent());
                                  },
                                  child: ListView.builder(
                                    itemBuilder: (context, index) =>
                                        children[index],
                                    itemCount: children.length,
                                  ),
                                ),
                              );
                            },
                          ),
                          floatingActionButton: FloatingActionButton(
                            onPressed: () {
                              Navigator.of(context)
                                  .pushNamed(ChatRouteType.chatCreateRoute);
                            },
                            child: Icon(Icons.add),
                          ),
                        );
                      }
                      if (state is ChatErrorState) {
                        return Center(
                          child: Text("${state.error}"),
                        );
                      }
                      return Center(
                        child: Text("Error"),
                      );
                    },
                  ),
                ),
              );
            }),
          ),
        ),
      ),
    );
  }

  @override
  void deactivate() {
    super.deactivate();
    widget.chatSocket.unsubscribePersonal();
  }

  @override
  void dispose() {
    widget.chatSocket.disconnect();
    super.dispose();
  }
}
