import 'dart:async';
import 'dart:convert';

import 'package:ocg_app/ocg_app.dart';
import 'package:ocg_chat/ocg_chat.dart';
import 'package:ocg_chat/src/models/socket.dart';
import 'package:ocg_chat/src/socket/base.dart';
import 'package:centrifuge/centrifuge.dart' as centrifuge;
import 'package:flutter/cupertino.dart';
import 'package:http/http.dart' as http;
import 'package:rxdart/subjects.dart';

class ChatSocket extends BaseSocket {
  final TokenRepository tokenRepository;
  final String url;
  final ChatRepository chatRepository;
  final MessageRepository messageRepository;
  static centrifuge.Client _client;
  final http.BaseClient httpClient;
  final BehaviorSubject<SocketChatData> _subject;

  static centrifuge.Subscription _personalSubscription;

  ChatSocket(this.url, this.tokenRepository, this.chatRepository,
      this.messageRepository, this.httpClient)
      : _subject = BehaviorSubject<SocketChatData>();

  @override
  Future<void> connect() async {
    debugPrint("Socket Connect ");
    if (_client == null) {
      _client = await createPrivateClient();
      var tokenResponse = await tokenRepository.getToken();
      if (!tokenResponse.isSuccess) {
        return null;
      }
      _client.setConnectData("\"${tokenResponse.data.token}\"".codeUnits);
      _client.connect();
    }
  }

  Future<centrifuge.Client> createPrivateClient() async {
    var tokenResponse = await tokenRepository.getToken();
    if (!tokenResponse.isSuccess) {
      return null;
    }
    AppToken token = tokenResponse.data;
    Map<String, dynamic> headers = {"Authorization": "Bearer ${token.token}"};
    centrifuge.Client client = centrifuge.createClient(
      url,
      config: centrifuge.ClientConfig(
          headers: headers..addAll({"Content-Type": 'application/json'}),
          onPrivateSub: (event) async {
            try {
              Map json = {
                "client": event.clientID,
                "channels": [event.channel]
              };
              var request = await httpClient.post('/centrifuge/subscribe',
                  body: jsonEncode(json), headers: ApiConfig.headers);
              var responseJson = jsonDecode(request.body);
              if (responseJson is Map) {
                var channelList = responseJson["channels"];
                if (channelList is List) {
                  String token = channelList[0]["token"];
                  return token;
                }
              }
              return null;
            } catch (e) {
              debugPrint("Error $e");
              return null;
            }
          }),
    );
    return client;
  }

  Future<void> subscribeToPersonalChat() async {
    debugPrint("Subscribe to Personal");
    if (_personalSubscription != null) {
      unsubscribePersonal();
    }
    AppResponse<AppToken> tokenResponse = await tokenRepository.getToken();
    if (!tokenResponse.isSuccess) {
      return;
    }
    String channel = "personal:#${tokenResponse.data.userId}";
    _personalSubscription = _client.getSubscription("$channel");
    _handleSubscription(_personalSubscription);
    _personalSubscription.subscribe();
  }

  void unsubscribePersonal() {
    if (_personalSubscription != null) {
      _personalSubscription.unsubscribe();
      _personalSubscription = null;
    }
  }

  void _handleSubscription(centrifuge.Subscription subscription) async {
    subscription.publishStream.listen((val) async {
      var json = jsonDecode(utf8.decode(val.data));
      SocketData data;
      if (json["type"] == SocketType.NEW_MSG) {
        SocketMessageData chatData =
            SocketMessageData.fromJson({"message": json});
        data = chatData.copyWith(
            message: (await messageRepository.decodeMessage(chatData.message)));
      } else if (json["type"] == SocketType.MSG_READ) {
        SocketReadData readData = SocketReadData.fromJson(json);
        data = readData;
      } else if (json["type"] == SocketType.MSG_DELETE) {
        SocketMsgDeleteData deleteData = SocketMsgDeleteData.fromJson(json);
        data = deleteData;
        // TODO Add
      } else if (json["type"] == SocketType.NEW_CHAT) {
        SocketNewChatData newChatData = SocketNewChatData.fromJson(json);
        data = newChatData;
        // TODO Add
      } else if (json["type"] == SocketType.CLEAR_CHAT) {
        SocketClearData clearData = SocketClearData.fromJson(json);
        data = clearData;
        // TODO Add
      } else if (json["type"] == SocketType.GROUP_INFO_CHANGED) {
        SocketGroupInfoChanged groupInfoChanged =
            SocketGroupInfoChanged.fromJson(json);
        data = groupInfoChanged;
        // TODO ADD
      }
      if (data != null) {
        _subject.add(data);
      }
    });
    subscription.joinStream.listen((val) {
      debugPrint("Socket Joined ${val.client} ${val.user}");
    });
    subscription.leaveStream.listen((val) {
      debugPrint("Socket Leave ${val.client} ${val.user}");
    });
    subscription.subscribeSuccessStream.listen((event) {
      debugPrint("ReSubscribed ${event.isResubscribed}");
    });
    subscription.unsubscribeStream.listen((val) {
      debugPrint("Unsubscribed");
    });
    subscription.subscribeErrorStream.listen((val) {
      debugPrint("Subscription Error ${val.message}");
    });
  }

  Stream<SocketData> get stream => _subject.stream;

  @override
  Future<void> disconnect() async {
    _subject.close();
    _client?.disconnect();
    _client = null;
  }
}
