import 'package:ocg_app/ocg_app.dart';
import 'package:ocg_chat/src/models/message.dart';
import 'package:flutter/material.dart';
import 'package:flutter_widgets/flutter_widgets.dart';

class MessageWidget extends StatelessWidget {
  final Message message;
  final void Function(Message message) onVisible;
  final void Function(Message message) onRetry;
  const MessageWidget({Key key, this.message, this.onVisible, this.onRetry})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return VisibilityDetector(
        key: Key("${message.id}"),
        child: GestureDetector(
            onLongPress: () async {
              if (message.isFailed) {
                String value = await AppUIHelper.of(context)
                    .showModalDialog(context, ["Retry"]);
                if (value == "Retry") {
                  onRetry(message);
                }
              }
            },
            child: buildContent()),
        onVisibilityChanged: (state) {
          if (state.visibleFraction > 0.9 && !message.isSelf) {
            if (onVisible != null) onVisible(message);
          }
        });
  }

  Widget buildContent() {
    Widget widget = Container();
    if (message.type == MessageType.text) {
      widget = buildText();
    }
    if (message.type == MessageType.image) {
      widget = buildImage();
    }

    return widget;
  }

  Widget buildWrapped({Widget child}) {
    return Container(
      margin: const EdgeInsets.only(top: 4, bottom: 4),
      padding: EdgeInsets.symmetric(horizontal: 12, vertical: 12),
      decoration: BoxDecoration(
          color: message.isSelf ? Color(0xff465ffd) : Color(0xfff4f5fa),
          borderRadius: BorderRadius.circular(20)),
      child: child,
    );
  }

  Widget buildText() {
    TextMessage textMessage = TextMessage.fromJson(message.content);
    return Builder(builder: (
      context,
    ) {
      return Container(
        alignment:
            message.isSelf ? Alignment.centerRight : Alignment.centerLeft,
        child: buildWrapped(
          child: Column(
            mainAxisSize: MainAxisSize.min,
            crossAxisAlignment: CrossAxisAlignment.end,
            children: <Widget>[
              Container(
                constraints: BoxConstraints(
                    minWidth: 100,
                    maxWidth: MediaQuery.of(context).size.width * 0.6),
                child: Container(
                  child: Column(
                    crossAxisAlignment: CrossAxisAlignment.start,
                    children: <Widget>[
                      Text(
                        "${textMessage.content}",
                        style: TextStyle(
                            color: message.isSelf
                                ? Color(0xffe4e9fe)
                                : Colors.black,
                            fontSize: 14.5,
                            height: 1.4),
                      ),
                    ],
                  ),
                ),
              ),
              SizedBox(
                height: 4,
              ),
              buildTime()
            ],
          ),
        ),
      );
    });
  }

  Widget buildImage() {
    ImageMessage imageMessage = ImageMessage.fromJson(message.content);
    return Container(
      alignment: message.isSelf ? Alignment.centerRight : Alignment.centerLeft,
      child: LayoutBuilder(builder: (context, constraints) {
        double size = constraints.biggest.width / 2;
        return buildWrapped(
          child: Container(
            width: size,
            child: Column(
              crossAxisAlignment: CrossAxisAlignment.end,
              mainAxisSize: MainAxisSize.min,
              children: <Widget>[
                Container(
                  width: size,
                  height: size * 1.4,
                  decoration: BoxDecoration(
                      borderRadius: BorderRadius.circular(8),
                      boxShadow: [
                        BoxShadow(
                            color: Colors.black12.withOpacity(0.07),
                            blurRadius: 3,
                            spreadRadius: 1)
                      ],
                      image: DecorationImage(
                          image: imageMessage.url != null
                              ? NetworkImage("${imageMessage.url}")
                              : NetworkImage("url"),
                          fit: BoxFit.cover)),
                  child: Center(),
                ),
                SizedBox(
                  height: 12,
                ),
                buildTime()
              ],
            ),
          ),
        );
      }),
    );
  }

  Widget buildTime({Color color}) {
    return Container(
      alignment: Alignment.centerRight,
      width: 90,
      child: Row(
        mainAxisSize: MainAxisSize.min,
        children: <Widget>[
          Text(
            "${DateHelper.getTimeString(message.recordedAt)}",
            style: TextStyle(
                color: color ??
                    (message.isSelf
                        ? Color(0xffdee3fe).withOpacity(0.8)
                        : Colors.grey),
                height: 1.4,
                fontSize: 12),
          ),
          message.isSelf
              ? message.isSent
                  ? Container(
                      margin: EdgeInsets.only(left: 4),
                      child: Icon(
                        (message.isSeen ?? false) ? Icons.done_all : Icons.done,
                        color: Color(0xffdee3fe).withOpacity(0.8),
                        size: 16,
                      ),
                    )
                  : Container(
                      margin: EdgeInsets.only(left: 4),
                      child: Icon(
                        message.isFailed ? Icons.error_outline : Icons.timer,
                        color: Color(0xffdee3fe).withOpacity(0.8),
                        size: 16,
                      ),
                    )
              : Container(
                  width: 0,
                  height: 0,
                )
        ],
      ),
    );
  }
}
